/***************************************************************************
    mmxcdplay.cpp  -  Class for playing CDs through a PCM audio device

    begin                : 5/26/2002
    copyright            : (C) 2002 by Michael Minn
    email                : michael@michaelminn.com

    This program is free software; you can redistribute it and/or modify
    it under the terms of version 2 of the GNU General Public License as 
    published by the Free Software Foundation (see COPYING or www.gnu.org).

    Subchannels:
    P-channel indicates the start and end of each track and was intended for
      simple audio players which did not have full Q-channel decoding.
    Q-channel contains the timecodes (minutes, seconds and frames),
      the TOC (in the lead-in), track type and catalogue number.
    Channels R to W are for subcode graphics (known as CD-G) and
      CD Text which accompany the main audio data.

***************************************************************************/

#include <wait.h>
#include <stdio.h>
#include <fcntl.h>
#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <pthread.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <linux/cdrom.h> // /usr/include/linux/cdrom.h
#include <linux/soundcard.h>

#include <Xm/Xm.h>
#include <Xm/List.h>
#include <Xm/Form.h>
#include <Xm/PushB.h>
#include <Xm/Label.h>
#include <Xm/ToggleB.h>
#include <Xm/MessageB.h>
#include <Xm/RowColumn.h>

#define MMXCDPLAY_VERSION_STRING "MMXCDPLAY v2002-11-29"
#define MMXCDPLAY_COPYRIGHT_STRING MMXCDPLAY_VERSION_STRING "(c) 2002 by Michael Minn"
#define MMXCDPLAY_PAGE_SIZE (CD_FRAMESIZE_RAW * CD_FRAMES)
#define MMXCDPLAY_BUFFER_PAGES 2

class mmxcdplay
{
	public:
	mmxcdplay(char *cdrom_device_name, char *audio_device_name);
	int mmxcdplay_errno(int error, char *message);
	int read_table_of_contents();
	int play_time(int seconds, int frame);
	int play_track(int track);
	int get_current_track();
	int get_track_seconds();
	int refresh_track_list();
	int read_thread();
	int write_thread();

	XtAppContext application;
	Widget shell;
	Widget main_widget;
	Widget time_display;
	Widget button_row;
	Widget play_button;
	Widget stop_button;
	Widget clear_button;
	Widget previous_button;
	Widget next_button;
	Widget autopause_button;
	Widget minus_15_button;
	Widget plus_15_button;
	Widget eject_button;
	Widget quit_button;
	Widget track_list;

	int cd_rom;
	int playing;
	int read_frame;
	int audio_device;
	int read_seconds;
	int write_seconds;
	pthread_t read_thread_id;
	pthread_t write_thread_id;
	struct cdrom_tochdr table_of_contents;
	struct cdrom_tocentry track_info[128];
	char message[256];
	char buffer[MMXCDPLAY_BUFFER_PAGES][MMXCDPLAY_PAGE_SIZE];
};


/****************************** Callbacks ******************************/

static void *create_read_thread(void *data)
{
	mmxcdplay *player = (mmxcdplay*) data;
	player->read_thread();
	return NULL;
}

static void *create_write_thread(void *data)
{
	mmxcdplay *player = (mmxcdplay*) data;
	player->write_thread();
	return NULL;
}

void button_callback(Widget widget, XtPointer client_data, XtPointer event_data)
{
	mmxcdplay *player = (mmxcdplay*) client_data;

	if ((widget == player->play_button) && !player->playing)
	{
		player->refresh_track_list();
		if (player->write_seconds < player->track_info[1].cdte_addr.msf.second)
			player->write_seconds = player->track_info[1].cdte_addr.msf.second;

		player->play_time(player->write_seconds, 0);
	}

	else if (widget == player->stop_button)
		player->playing = 0;

	else if (widget == player->clear_button)
	{
		player->playing = 0;
		player->read_seconds = 0;
		player->write_seconds = 0;
		player->refresh_track_list();
	}

	else if (widget == player->previous_button)
	{
		if (player->get_current_track() > 0)
			player->play_track(player->get_current_track() - 1);
	}

	else if (widget == player->next_button)
		player->play_track(player->get_current_track() + 1);

	else if (widget == player->minus_15_button)
	{
		if (player->write_seconds > 17)
		{
			int new_seconds = player->write_seconds - 15;
			player->read_seconds = new_seconds;
			player->write_seconds = new_seconds;
		}
	}

	else if (widget == player->plus_15_button)
	{
		int new_seconds = player->write_seconds + 15;
		player->read_seconds = new_seconds;
		player->write_seconds = new_seconds;
	}

	else if (widget == player->eject_button)
	{
		player->playing = 0;
		player->read_seconds = 0;
		player->write_seconds = 0;
		sleep(1);
		ioctl(player->cd_rom, CDROMEJECT, 0);
		XmListDeleteAllItems(player->track_list);
	}

	else if (widget == player->quit_button)
		exit(0);
}

static void update_time(XtPointer client_data, XtIntervalId *interval)
{
	mmxcdplay *player = (mmxcdplay*) client_data;

	int track = player->get_current_track();
	int total = player->write_seconds;
	int seconds = player->get_track_seconds();
	char buffer[128] = "";

	if (track > 0)
	{
		sprintf(buffer, "%d - %02d:%02d %02d:%02d", track, seconds / 60, seconds % 60, total / 60, total % 60);
		XmListSelectPos(player->track_list, track, 0);
	}

	XmString label = XmStringCreateSimple(buffer);
	XtVaSetValues(player->time_display, XmNlabelString, label, NULL);
	XmStringFree(label);

	if (player->message[0])
	{
		XmString caption_string = XmStringCreateLocalized("Error");
		XmString message_string = XmStringCreateLocalized(player->message);
		Widget dialog = XmCreateErrorDialog(player->shell, "errno", NULL, 0);
		XtVaSetValues(dialog, 
			XmNdialogTitle, caption_string, 
			XmNmessageString, message_string, NULL);
		XmStringFree(caption_string);
		XmStringFree(message_string);
		XtUnmanageChild(XtNameToWidget(dialog, "Cancel"));
		XtUnmanageChild(XtNameToWidget(dialog, "Help"));
		XtManageChild(dialog);
		player->message[0] = 0;
	}

	XtAppAddTimeOut(player->application, 250, update_time, player);
}

static void track_list_select(Widget widget, XtPointer client_data, XtPointer call_data)
{
	int *track = NULL;
	int track_count = 0;
	mmxcdplay *player = (mmxcdplay*) client_data;

	XmListGetSelectedPos(player->track_list, &track, &track_count);

	// XtVaGetValues(track_list, XmNselectedPositions, &track, NULL);

	if (track)
	{
		player->play_track(*track);
		XtFree((char*)track);
	}
}


/****************************** Class Functions ******************************/

mmxcdplay::mmxcdplay(char *cdrom_device_name, char *audio_device_name)
{
	int fake_argc = 1;
	char *program_name = "mmxcdplay";
	String resources[] = { "mmxcdplay*background: white", "mmxcdplay*foreground: black", NULL };

	memset(this, 0, sizeof(*this));

	shell = XtVaOpenApplication(&application, "mmxcdplay", NULL, 0,
		&fake_argc, &program_name, resources, sessionShellWidgetClass, 
		XmNwidth, 270,
		XmNtitle, MMXCDPLAY_VERSION_STRING, 
		NULL);

	main_widget = XtVaCreateManagedWidget("form", xmRowColumnWidgetClass, shell,
		XmNorientation, XmVERTICAL,
		NULL);

	time_display = XtVaCreateManagedWidget("time", xmLabelWidgetClass, main_widget,
		XmNlabelType, XmSTRING,
        	NULL);

	button_row = XtVaCreateManagedWidget("button_row1", xmRowColumnWidgetClass, main_widget,
		XmNorientation, XmHORIZONTAL,
		XmNentryAlignment, XmALIGNMENT_CENTER,
		XmNnumColumns, 4,
		XmNpacking, XmPACK_COLUMN,
		NULL);

	play_button = XtVaCreateManagedWidget("play", xmPushButtonWidgetClass, button_row,
        	XmNlabelString, XmStringCreateSimple("Play"), NULL);
	XtAddCallback(play_button, XmNactivateCallback, button_callback, (XtPointer)this);

	stop_button = XtVaCreateManagedWidget("stop", xmPushButtonWidgetClass, button_row,
        	XmNlabelString, XmStringCreateSimple("Stop"), NULL);
	XtAddCallback(stop_button, XmNactivateCallback, button_callback, (XtPointer)this);

	clear_button = XtVaCreateManagedWidget("clear", xmPushButtonWidgetClass, button_row,
        	XmNlabelString, XmStringCreateSimple("Clear"), NULL);
	XtAddCallback(clear_button, XmNactivateCallback, button_callback, (XtPointer)this);

	previous_button = XtVaCreateManagedWidget("previous", xmPushButtonWidgetClass, button_row,
        	XmNlabelString, XmStringCreateSimple("Previous"), NULL);
	XtAddCallback(previous_button, XmNactivateCallback, button_callback, (XtPointer)this);

	next_button = XtVaCreateManagedWidget("next", xmPushButtonWidgetClass, button_row,
        	XmNlabelString, XmStringCreateSimple("Next"), NULL);
	XtAddCallback(next_button, XmNactivateCallback, button_callback, (XtPointer)this);

	autopause_button = XtVaCreateManagedWidget("autopause", xmToggleButtonWidgetClass, button_row,
        	XmNlabelString, XmStringCreateSimple("Autopause"), NULL);

	minus_15_button = XtVaCreateManagedWidget("minus_15", xmPushButtonWidgetClass, button_row,
        	XmNlabelString, XmStringCreateSimple("-0:15"), NULL);
	XtAddCallback(minus_15_button, XmNactivateCallback, button_callback, (XtPointer)this);

	plus_15_button = XtVaCreateManagedWidget("play_15", xmPushButtonWidgetClass, button_row,
        	XmNlabelString, XmStringCreateSimple("+0:15"), NULL);
	XtAddCallback(plus_15_button, XmNactivateCallback, button_callback, (XtPointer)this);

	eject_button = XtVaCreateManagedWidget("eject", xmPushButtonWidgetClass, button_row,
        	XmNlabelString, XmStringCreateSimple("Eject"), NULL);
	XtAddCallback(eject_button, XmNactivateCallback, button_callback, (XtPointer)this);

	quit_button = XtVaCreateManagedWidget("quit", xmPushButtonWidgetClass, button_row,
        	XmNlabelString, XmStringCreateSimple("Quit"), NULL);
	XtAddCallback(quit_button, XmNactivateCallback, button_callback, (XtPointer)this);

	track_list = XmCreateScrolledList(main_widget, "track_list", NULL, 0);
	XtVaSetValues(track_list,
		XmNselectionPolicy, XmSINGLE_SELECT,
		XmNwidth, 200,
		XmNvisibleItemCount, 10,
		NULL);
	XtManageChild(track_list);

	XtAddCallback(track_list, XmNdefaultActionCallback, track_list_select, (XtPointer)this);
	XtAddCallback(track_list, XmNsingleSelectionCallback, track_list_select, (XtPointer)this);

	XtRealizeWidget(shell);

	XtAppAddTimeOut(application, 250, update_time, this);

	int channels = 2;
	int sampling_rate = 44100;
	int audio_format = AFMT_S16_LE;

	if ((audio_device = open(audio_device_name, O_WRONLY)) <= 0)
		mmxcdplay_errno(errno, audio_device_name);

	else if (ioctl(audio_device, SNDCTL_DSP_SETFMT, &audio_format) < 0)
		mmxcdplay_errno(errno, "Failure setting audio device format");

	else if (ioctl(audio_device, SNDCTL_DSP_CHANNELS, &channels) < 0)
		mmxcdplay_errno(errno, "Failure setting audio device channels");

	else if (ioctl(audio_device, SNDCTL_DSP_SPEED, &sampling_rate) < 0)
		mmxcdplay_errno(errno, "Failure setting audio device sampling rate");

	else if ((cd_rom = open(cdrom_device_name, O_RDONLY | O_NONBLOCK)) <= 0)
		mmxcdplay_errno(errno, cdrom_device_name);

	else if (refresh_track_list() < 0)
		;

	else if (pthread_create(&read_thread_id, NULL, &create_read_thread, this) != 0)
		mmxcdplay_errno(errno, "Failure creating read process thread");

	if (pthread_create(&write_thread_id, NULL, &create_write_thread, this) != 0)
		mmxcdplay_errno(errno, "Failure creating read process thread");
}

int mmxcdplay::mmxcdplay_errno(int error, char *message)
{
	sprintf(this->message, "%s: %s", message, strerror(errno));
	puts(this->message);
	return -error;
}

int mmxcdplay::read_table_of_contents()
{
	if (cd_rom < 0)
		return mmxcdplay_errno(EINVAL, "CD-ROM not open");

	if (ioctl(cd_rom, CDROMREADTOCHDR, &table_of_contents) < 0)
		return mmxcdplay_errno(errno, "Failure reading table of contents header - may not be an audio CD");

	for (int track = table_of_contents.cdth_trk0; track <= table_of_contents.cdth_trk1; ++track)
	{
		track_info[track].cdte_track = track;
		track_info[track].cdte_format = CDROM_MSF;
		if (ioctl(cd_rom, CDROMREADTOCENTRY, &track_info[track]) < 0)
			return mmxcdplay_errno(errno, "Failure reading table of contents track info");
	}

	return table_of_contents.cdth_trk1;
}

int mmxcdplay::play_time(int seconds, int frame)
{
	playing = 0;

	read_seconds = seconds;
	read_frame = frame;
	write_seconds = seconds;

	playing = 1;
	return 0;
}

int mmxcdplay::play_track(int track)
{
	if (read_table_of_contents() < 0)
		return -errno;

	if ((track < table_of_contents.cdth_trk0) || (track > table_of_contents.cdth_trk1))
		return mmxcdplay_errno(EINVAL, "Invalid track number");

	return play_time((track_info[track].cdte_addr.msf.minute * 60) + track_info[track].cdte_addr.msf.second, track_info[track].cdte_addr.msf.frame);
}

int mmxcdplay::read_thread()
{
	while (1)
	{
		pthread_testcancel();
		while (!playing)
		{
			usleep(10000);
			pthread_testcancel();
		}

		struct cdrom_read_audio audio_data;
		audio_data.addr.msf.minute = read_seconds / 60;
		audio_data.addr.msf.second = read_seconds % 60;
		audio_data.addr.msf.frame = read_frame;
		audio_data.addr_format = CDROM_MSF;
		audio_data.nframes = CD_FRAMES;
		audio_data.buf = (__u8*) &buffer[read_seconds % MMXCDPLAY_BUFFER_PAGES];

		if (ioctl(cd_rom, CDROMREADAUDIO, &audio_data) < 0)
			if ((get_current_track() == table_of_contents.cdth_trk1) && (errno == EIO))
			{
				printf("End of CD\n");
				playing = 0;
			}
			else
			{
				mmxcdplay_errno(errno, "Failure reading audio data");
				playing = 0;
			}

		if (playing)
			++read_seconds;

		while (playing && ((read_seconds - MMXCDPLAY_BUFFER_PAGES) >= write_seconds))
		{
			usleep(10000);
			pthread_testcancel();
		}
	}

	return 0;
}

int mmxcdplay::write_thread()
{
	while (1)
	{
		pthread_testcancel();
		while (!playing)
		{
			usleep(10000);
			pthread_testcancel();
		}

		while (playing && (write_seconds >= read_seconds))
		{
			usleep(10000);
			pthread_testcancel();
		}

		if (write(audio_device, &buffer[write_seconds % MMXCDPLAY_BUFFER_PAGES], MMXCDPLAY_PAGE_SIZE) < MMXCDPLAY_PAGE_SIZE)
		{
			mmxcdplay_errno(errno, "Failure writing audio data");
			playing = 0;
		}

		if (playing)
		{
			int old_track = get_current_track();
			++write_seconds;
			if (XmToggleButtonGetState(autopause_button) && (old_track != get_current_track()))
				playing = 0;
		}
	}

	return 0;
}

int mmxcdplay::get_current_track()
{
	for (int track = table_of_contents.cdth_trk0; track <= table_of_contents.cdth_trk1; ++track)
		if (write_seconds < ((track_info[track].cdte_addr.msf.minute * 60) + track_info[track].cdte_addr.msf.second))
			return track - 1;

	return table_of_contents.cdth_trk1;
}

int mmxcdplay::get_track_seconds()
{
	int track = get_current_track();
	int track_seconds = (track_info[track].cdte_addr.msf.minute * 60) + track_info[track].cdte_addr.msf.second;
	return write_seconds - track_seconds;
}

int mmxcdplay::refresh_track_list()
{
	XmListDeleteAllItems(track_list);

	read_table_of_contents();

	for (int track = table_of_contents.cdth_trk0; track <= table_of_contents.cdth_trk1; ++track)
	{
		char string[128];
		sprintf(string, "%d - %02d:%02d:%02d",
			track_info[track].cdte_track,
			track_info[track].cdte_addr.msf.minute,
			track_info[track].cdte_addr.msf.second,
			track_info[track].cdte_addr.msf.frame);

		XmString label = XmStringCreateSimple(string);
		XmListAddItem(track_list, label, 0);
		XmStringFree(label);
	}

	return 0;
}

int main(int argc, char **argv)
{
	printf(MMXCDPLAY_COPYRIGHT_STRING "\n");
	if (argc < 3)
	{
		printf("SYNTAX: mmxcdplay <cd device> <audio device>\n");
		printf("        example: mmxcdplay /dev/cdrom /dev/dsp1\n");
		return 0;
	}

	mmxcdplay player(argv[1], argv[2]);

	XtAppMainLoop(player.application);

	return 0;
}













